#include <neopixel.h>
#include <math.h>
#include "application.h"

#define PIXEL_PIN D3
#define BACKLIGHT_PIN D6
#define PIXEL_TYPE WS2812B
#define PIXEL_COUNT 60

#define PATTERN_FLAGS(palette, len) (palette << 28) | len

Adafruit_NeoPixel strip = Adafruit_NeoPixel(PIXEL_COUNT, PIXEL_PIN, PIXEL_TYPE);
Adafruit_NeoPixel backlight = Adafruit_NeoPixel(PIXEL_COUNT, BACKLIGHT_PIN, PIXEL_TYPE);

typedef struct {
  uint32_t flags;
  char* name;
  uint32_t* colors;
} pattern;

const PROGMEM uint32_t rgb4[1] = { 0x1b1b1b1b };
const PROGMEM uint32_t rgb8[1] = { 01234567 << 8 };
const PROGMEM uint32_t rgb16[6] = { 0x1234567, 0x89abcdef };
const PROGMEM uint32_t blues[4] = { 0xa0a2a6ae, 0xb2b6b8ba, 0xbab8b6b2, 0xaea6a2a0 };
const PROGMEM pattern rgb4Pattern = { PATTERN_FLAGS(2u, 16), "rainbow low", ((uint32_t*)(rgb4)) };
const PROGMEM pattern rgb8Pattern = { PATTERN_FLAGS(3u, 8), "rainbow medium", ((uint32_t*)(rgb8)) };
const PROGMEM pattern rgb16Pattern = { PATTERN_FLAGS(4u, 16), "rainbow high", ((uint32_t*)(rgb16)) };
const PROGMEM pattern bluesPattern = { PATTERN_FLAGS(8u, 16), "blues", ((uint32_t*)(blues)) };

const PROGMEM pattern patterns[] = {
  rgb4Pattern,
  rgb8Pattern,
  rgb16Pattern,
  bluesPattern,
 };

const int d = 16;
const int STEP_TIME = 100;

retained char pattern_name[256] = { 0 };
int pattern_index = 0;
int offset = 0;

int set_pattern_by_index(String index);
int set_pattern_by_name(String name);

void setup() {
  strip.begin();
  strip.show();
  backlight.begin();
  for (int i = 0; i < PIXEL_COUNT; i++) {
    backlight.setPixelColor(i, 0x606060);
  }
  backlight.show();
  set_pattern_by_name(pattern_name);
  Particle.function("set_index", set_pattern_by_index);
  Particle.function("set_name", set_pattern_by_name);
}

uint32_t hue(byte v) {
    if(v < 85) {
        return strip.Color(
          (255 - v * 3) / d,
          (v * 3) / d,
          0 / d
        );
    } else if(v < 170) {
        v -= 85;
        return strip.Color(
          0 / d,
          (255 - v * 3) / d,
          (v * 3) / d
        );
    } else {
        v -= 170;
        return strip.Color(
          (v * 3) / d,
          0 / d,
          (255 - v * 3) / d
        );
    }
}

void loop() {
  const pattern pattern = patterns[pattern_index];
  int32_t b = pattern.flags >> 28;
  int32_t palette_size = 1 << b;
  int32_t pattern_length = pattern.flags & 0x0fffffff;
  if (palette_size == 0 || pattern_length == 0) {
    delay(STEP_TIME);
    return;
  }
  if (pattern_length > 0) {
    for (int32_t i = 0; i < PIXEL_COUNT; i++) {
      uint32_t pattern_idx = (i + offset) % pattern_length;
      uint32_t array_idx = pattern_idx / (32 / b);

      uint64_t low = pattern.colors[array_idx];
      uint64_t high = pattern.colors[array_idx - 1];
      uint64_t whole = (high << 32) | low;

      uint32_t amt_shifted = (array_idx + 1) * 32;
      uint32_t bit_offset = (pattern_idx + 1) * b;

      uint32_t color = (whole >> (amt_shifted - bit_offset)) & ((1 << b) - 1);

      strip.setPixelColor(i, hue(color * 256 / palette_size));
    }
  }
  strip.show();
  offset++;
  delay(STEP_TIME);
}

int set_pattern_by_index(String index) {
  pattern_index = atoi(index);
  offset = 0;
  return 1;
}

int set_pattern_by_name(String name) {
  name.toCharArray(pattern_name, sizeof(pattern_name));

  for (int i = 0; i < sizeof(patterns) / sizeof(pattern); i++) {
    if (name.equals(patterns[i].name)) {
      pattern_index = i;
      return 1;
    }
  }
  return -1;
}

void publish_temp(double temp) { // TODO: connect temp sensor
  Particle.publish("temp", String(temp), 60000, PRIVATE);
}
